;; Generated package description from wcheck-mode.el  -*- no-byte-compile: t -*-
(define-package "wcheck-mode" "2021.0.20220101.81620" "General interface for text checkers" 'nil :commit "4a2439ed486342d825259b62cde2e73ef34dbc04" :authors '(("Teemu Likonen" . "tlikonen@iki.fi")) :maintainer '("Teemu Likonen" . "tlikonen@iki.fi") :keywords '("text" "spell" "check" "languages" "ispell") :url "https://github.com/tlikonen/wcheck-mode")
